use crate::prelude::*;

use std::os::raw::c_void;
use std::ptr;

#[repr(C)]
pub struct VkAccelerationStructureBuildGeometryInfoKHR {
    pub sType: VkStructureType,
    pub pNext: *const c_void,
    pub r#type: VkAccelerationStructureTypeKHR,
    pub flags: VkBuildAccelerationStructureFlagBitsKHR,
    pub mode: VkBuildAccelerationStructureModeKHR,
    pub srcAccelerationStructure: VkAccelerationStructureKHR,
    pub dstAccelerationStructure: VkAccelerationStructureKHR,
    pub geometryCount: u32,
    pub pGeometries: *const VkAccelerationStructureGeometryKHR,
    pub ppGeometries: *const *const VkAccelerationStructureGeometryKHR,
    pub scratchData: VkDeviceOrHostAddressKHR,
}

impl VkAccelerationStructureBuildGeometryInfoKHR {
    pub fn new(
        r#type: VkAccelerationStructureTypeKHR,
        flags: impl Into<VkBuildAccelerationStructureFlagBitsKHR>,
        mode: VkBuildAccelerationStructureModeKHR,
        src_acceleration_structure: VkAccelerationStructureKHR,
        dst_acceleration_structure: VkAccelerationStructureKHR,
        scratch_data: VkDeviceOrHostAddressKHR,
    ) -> Self {
        VkAccelerationStructureBuildGeometryInfoKHR {
            sType: VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_BUILD_GEOMETRY_INFO_KHR,
            pNext: ptr::null(),
            r#type,
            flags: flags.into(),
            mode,
            srcAccelerationStructure: src_acceleration_structure,
            dstAccelerationStructure: dst_acceleration_structure,
            geometryCount: 0,
            pGeometries: ptr::null(),
            ppGeometries: ptr::null(),
            scratchData: scratch_data,
        }
    }

    pub fn minimal(
        r#type: VkAccelerationStructureTypeKHR,
        flags: impl Into<VkBuildAccelerationStructureFlagBitsKHR>,
        mode: VkBuildAccelerationStructureModeKHR,
    ) -> Self {
        Self::new(
            r#type,
            flags,
            mode,
            VkAccelerationStructureKHR::NULL_HANDLE,
            VkAccelerationStructureKHR::NULL_HANDLE,
            VkDeviceOrHostAddressKHR::null(),
        )
    }
}

pub trait SetGeometry<T> {
    fn set_geometry(&mut self, geometries: T);
}

impl<'a> SetGeometry<&'a [&'a VkAccelerationStructureGeometryKHR]>
    for VkAccelerationStructureBuildGeometryInfoKHR
{
    fn set_geometry(&mut self, geometries: &'a [&'a VkAccelerationStructureGeometryKHR]) {
        debug_assert_eq!(self.pGeometries, ptr::null());

        self.geometryCount = geometries.len() as u32;
        self.ppGeometries = geometries.as_ptr() as *const *const _;
    }
}

impl<'a> SetGeometry<&'a [VkAccelerationStructureGeometryKHR]>
    for VkAccelerationStructureBuildGeometryInfoKHR
{
    fn set_geometry(&mut self, geometries: &'a [VkAccelerationStructureGeometryKHR]) {
        debug_assert_eq!(self.ppGeometries, ptr::null());

        self.geometryCount = geometries.len() as u32;
        self.pGeometries = geometries.as_ptr() as *const _;
    }
}
